import pathlib, pandas, datetime, subprocess, os, logging,subprocess,collections,json
from abritamr.version import db
from abritamr.CustomLog import CustomFormatter


class Setup(object):
    """
    A base class for setting up abritamr return a valid input object for subsequent steps
    """
    def __init__(self, args):
        

        self.logger =logging.getLogger(__name__) 
        self.logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(CustomFormatter())
        fh = logging.FileHandler('abritamr.log')
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
        fh.setFormatter(formatter)
        self.logger.addHandler(ch) 
        self.logger.addHandler(fh)
        
        
    def file_present(self, name):
        """
        check file is present
        :name is the path of the file to check
        """
        
        if name == "":
            return False
        elif pathlib.Path(name).exists():
            self.logger.info(f"Checking if file {name} exists")
            return True
        else:
            return False

class SetupAMR(Setup):

    """
    setup amr inputs for amrfinder run
    """
    def __init__(self, args):
        

        # for amr
        self.species_list = json.load(open(f"{pathlib.Path(__file__).parent / 'species_config.json'}",'r'))
        
        self.logger =logging.getLogger(__name__) 
        self.logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(CustomFormatter())
        fh = logging.FileHandler('abritamr.log')
        fh.setLevel(logging.INFO)
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
        fh.setFormatter(formatter)
        self.logger.addHandler(ch) 
        self.logger.addHandler(fh)

        
        self.jobs = args.jobs # number of amrfinderplus to run at a time
        self.contigs = args.contigs
        self.from_contigs = True if args.contigs != '' else False
        self.prefix = args.prefix
        # amr
        self.species = args.species if args.species in self.species_list else ""
        self.identity = args.identity
        self.amrfinder_db = args.amrfinder_db

        

    def _check_prefix(self):
        """
        If running type is not batch, then check that prefix is present and a string
        """

        if self.prefix == '':
            self.logger.critical(f"You must supply a sample or sequence id.")
            raise SystemExit
        else:
            return True

            
    
    def _get_input_shape(self):
        """
        determine shape of file
        """
        run_type = 'assembly'
        with open(self.contigs, 'r') as c:
            data = c.read().strip().split('\n')
            firstline = data[0]
            if not firstline.startswith('>'):
                for line in data:
                    if len(line.split('\t')) != 2:
                        self.logger.critical("Your input file should either be a tab delimited file with two columns or the path to contigs. Please check your input and try again.")
                        raise SystemExit
                run_type = 'batch'
        self.logger.info(f"The input file seems to be in the correct format. Thank you.")
        return run_type
    

    def _input_files(self):
        """
        Ensure that the files (either contigs or amrfinder output) exist and return running type
        """
        
        running_type = self._get_input_shape()
        if running_type == 'batch':
            self.logger.info(f"Checking that the input data is present.")
            with open(self.contigs, 'r') as c:
                data = c.read().strip().split('\n')
                for line in data:
                    row = line.split('\t')
                    if not self.file_present(row[1]):
                        self.logger.critical(f"{row[1]} is not a valid file path. Please check your input and try again.")
                        raise SystemExit
        elif running_type == 'assembly' and self.file_present(self.contigs):
            self.logger.info(f"{self.contigs} is present. abritamr can proceed.")
        else:
            self.logger.critical(f"Something has gone wrong with your inputs. Please try again.")
            raise SystemExit
        
        return running_type
   

    def setup(self):
        # check that inputs are correct and files are present
        running_type = self._input_files()
        # check that prefix is present (if needed)
        if running_type == 'assembly':
            self._check_prefix()
        
        Data = collections.namedtuple('Data', ['run_type', 'input', 'prefix', 'jobs', 'organism', 'identity','amrfinder_db'])
        input_data = Data(running_type, self.contigs, self.prefix, self.jobs, self.species, self.identity, self.amrfinder_db)
        
        return input_data


class SetupMDU(Setup):
    """
    Setup MDUify of abritamr results
    """
    def __init__(self, args):
        

        self.logger =logging.getLogger(__name__) 
        self.logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(CustomFormatter())
        fh = logging.FileHandler('abritamr.log')
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
        fh.setFormatter(formatter)
        self.logger.addHandler(ch) 
        self.logger.addHandler(fh)
        self.db = db
        self.qc = args.qc
        self.runid = args.runid
        self.matches = args.matches
        self.partials = args.partials   
        self.sop = args.sop
        self.sop_name = args.sop_name

    def _check_runid(self):
        if self.runid == '':
            self.logger.critical(f"Run ID can not be empty, please try again.")
            raise SystemExit
        else:
            return True

    def setup(self):
        """
        Check the inputs for MDU - ensure all files are present for collation.
        """
        
        file_dict = {
            'QC': self.qc, 
            'summary_matches': self.matches,
            'summary_partials':self.partials
            } if self.sop == 'general' else {
            'QC': self.qc, 
            'summary_matches': self.matches,
            # 'summary_partials':self.partials
            }

        if self._check_runid():
            self.logger.info(f"You are generating a {'general report' if self.sop == 'general' else 'species specific report'}")
            self.logger.info(f"Now checking all input files are present.")
            for _file in file_dict:
                self.logger.info(f"Checking that {_file} is present.")
                if self.file_present(file_dict[_file]):
                    self.logger.info(f"{file_dict[_file]} is present.")
                else:
                    self.logger.critical(f"The {_file} file supplied ({file_dict[_file]}) does not exist. Please check your inputs and try again.")
                    raise SystemExit

            Data = collections.namedtuple('Data', ['qc', 'matches', 'partials', 'db', 'runid', 'sop','sop_name'])
        
            return Data(self.qc, self.matches, self.partials, self.db, self.runid, self.sop, self.sop_name)
        


import sys, pathlib, pandas, pytest, numpy, logging, collections

from unittest.mock import patch, PropertyMock

from abritamr.AmrSetup import Setup, SetupAMR, SetupMDU
from abritamr.RunFinder import RunFinder
from abritamr.Collate import Collate, MduCollate



test_folder = pathlib.Path(__file__).parent
REFGENES = f"{pathlib.Path(__file__).parent.parent /'abritamr' /'db' / 'refgenes_latest.csv'}"
CONTROLS = pathlib.Path(__file__).parent.parent /'abritamr' / 'control'

def test_file_present():
    """
    assert true when the input file is true
    """
    with patch.object(Setup, "__init__", lambda x: None):
        amr_obj = Setup()
        amr_obj.logger = logging.getLogger(__name__)
        p = CONTROLS / "contigs.fa"
    
        assert amr_obj.file_present(p)

# Test SetupAMR
def test_check_run_type_batch():
    """
    assert true when a tab file is provided with two columns
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{test_folder / 'batch.txt'}"
        amr_obj.prefix = ''
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._get_input_shape() == 'batch'

def test_check_run_type_single():
    """
    assert true when a contigs file is provided and batch recorded as assembly
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{CONTROLS / 'contigs.fa'}"
        amr_obj.prefix = 'somename'
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._get_input_shape() == 'assembly'

def test_check_run_type_wrong_data():
    """
    assert true when error raised because input is wrong shape
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{test_folder / 'batch_fail.txt'}"
        amr_obj.prefix = ''
        amr_obj.logger = logging.getLogger(__name__)
        with pytest.raises(SystemExit):
            amr_obj._get_input_shape()

def test_prefix_string():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.prefix = "some_isolate"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._check_prefix()

def test_prefix_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.prefix = ""
        amr_obj.logger = logging.getLogger(__name__)
        with pytest.raises(SystemExit):
            amr_obj._check_prefix()
       

def test_setup_contigs():
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{CONTROLS / 'contigs.fa'}"
        amr_obj.prefix = 'somename'
        amr_obj.jobs  = 16
        amr_obj.species = ''
        amr_obj.identity = ''
        amr_obj.amrfinder_db = f"{pathlib.Path(__file__).parent.parent /'abritamr' /'db' / 'amrfinderplus'/ 'data'/ '2022-08-09.1'}"
        amr_obj.logger = logging.getLogger(__name__)
        T = collections.namedtuple('T', ['run_type', 'input', 'prefix', 'jobs', 'organism', 'identity','amrfinder_db'])
        input_data = T('assembly', amr_obj.contigs, amr_obj.prefix, amr_obj.jobs, amr_obj.species, amr_obj.identity, amr_obj.amrfinder_db)
        assert amr_obj.setup() == input_data

def test_species():
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{CONTROLS / 'contigs.fa'}"
        amr_obj.prefix = 'somename'
        amr_obj.jobs  = 16
        amr_obj.species = 'Neiserria'
        amr_obj.identity = ''
        amr_obj.amrfinder_db = f"{pathlib.Path(__file__).parent.parent /'abritamr' /'db' / 'amrfinderplus'/ 'data'/ '2022-08-09.1'}"
        amr_obj.logger = logging.getLogger(__name__)
        T = collections.namedtuple('T', ['run_type', 'input', 'prefix', 'jobs', 'organism', 'identity','amrfinder_db'])
        input_data = T('assembly', amr_obj.contigs, amr_obj.prefix, amr_obj.jobs, amr_obj.species, amr_obj.identity, amr_obj.amrfinder_db)
        assert amr_obj.setup() == input_data


def test_setup_batch():
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{test_folder / 'batch.txt'}"
        amr_obj.prefix = ''
        amr_obj.jobs  = 16
        amr_obj.species = ''
        amr_obj.identity = ''
        amr_obj.amrfinder_db = f"{pathlib.Path(__file__).parent.parent /'abritamr' /'db' / 'amrfinderplus'/ 'data'/ '2022-08-09.1'}"
        amr_obj.logger = logging.getLogger(__name__)
        T = collections.namedtuple('T', ['run_type', 'input', 'prefix', 'jobs', 'organism','identity', 'amrfinder_db'])
        input_data = T('batch', amr_obj.contigs, amr_obj.prefix, amr_obj.jobs, amr_obj.species, amr_obj.identity,amr_obj.amrfinder_db)
        assert amr_obj.setup() == input_data
 
def test_setup_fail():
    with patch.object(SetupAMR, "__init__", lambda x: None):
        amr_obj = SetupAMR()
        amr_obj.contigs = f"{test_folder / 'batch_fail.txt'}"
        amr_obj.prefix = ''
        amr_obj.jobs  = 16
        amr_obj.species = ''

        amr_obj.logger = logging.getLogger(__name__)
        T = collections.namedtuple('T', ['run_type', 'input', 'prefix', 'jobs', 'organism'])
        input_data = T('batch', amr_obj.contigs, amr_obj.prefix, amr_obj.jobs, amr_obj.species)
        with pytest.raises(SystemExit):
            amr_obj.setup()

# # Test SetupMDU
MDU = collections.namedtuple('MDU', ['runid', 'matches', 'partials', 'qc', 'sop', 'sop_name'])

def test_prefix_string():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        args = MDU("RUNID", 'tests/summary_matches.txt', 'tests/summary_matches.txt', 'tests/mdu_qc_checked.csv', 'general', 'sop_name')
        amr_obj = SetupMDU(args)
        # amr_obj.runid= 
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._check_runid()

def test_prefix_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        args = MDU("", 'tests/summary_matches.txt', 'tests/summary_matches.txt', 'tests/mdu_qc_checked.csv', 'general', 'sop_name')
        amr_obj = SetupMDU(args)
        amr_obj.logger = logging.getLogger(__name__)
        with pytest.raises(SystemExit):
            amr_obj._check_runid()


def test_mdu_setup_success():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        args = MDU("RUNID", 'tests/summary_matches.txt', 'tests/summary_matches.txt', 'tests/mdu_qc_checked.csv', 'general', 'sop_name')
        amr_obj = SetupMDU(args)
        Data = collections.namedtuple('Data', ['qc', 'matches', 'partials', 'db', 'runid','sop', 'sop_name'])
        d = Data(args.qc, args.matches, args.partials, amr_obj.db, args.runid, args.sop, args.sop_name)
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.setup() == d


def test_mdu_setup_fail():
    """
    assert True when non-empty string is given
    """
    with patch.object(SetupAMR, "__init__", lambda x: None):
        args = MDU("RUNID", 'tests/summarymatches.txt', 'tests/summary_matches.txt', 'tests/mdu_qc_checked.csv', 'general', 'sop_name')
        amr_obj = SetupMDU(args)
        Data = collections.namedtuple('Data', ['qc', 'matches', 'partials', 'db', 'runid'])
        d = Data(args.qc, args.matches, args.partials, amr_obj.db, args.runid)
        amr_obj.logger = logging.getLogger(__name__)
        with pytest.raises(SystemExit):
            amr_obj.setup()

# # test RunFinder

Data = collections.namedtuple('Data', ['run_type', 'input', 'prefix', 'jobs', 'organism'])
def test_batch_cmd_no_org():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        
        args = Data("batch", 'tests/batch.txt', '', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"parallel -j {args.jobs} --colsep '\\t' 'mkdir -p {{1}} && amrfinder -n {{2}} -o {{1}}/amrfinder.out --plus  --threads 1 -d {amr_obj.amrfinder_db}' :::: {args.input}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._batch_cmd() == cmd

def test_batch_cmd_with_org():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/batch.txt', '', 9, 'Salmonella')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"parallel -j {args.jobs} --colsep '\\t' 'mkdir -p {{1}} && amrfinder -n {{2}} -o {{1}}/amrfinder.out --plus --organism {args.organism} --threads 1 -d {amr_obj.amrfinder_db}' :::: {args.input}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._batch_cmd() == cmd

def test_batch_cmd_with_neiserria():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/batch.txt', '', 9, 'Neiserria')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"parallel -j {args.jobs} --colsep '\\t' 'mkdir -p {{1}} && amrfinder -n {{2}} -o {{1}}/amrfinder.out --plus --organism {args.organism} --threads 1 -d {amr_obj.amrfinder_db}' :::: {args.input}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._batch_cmd() == cmd

def test_single_cmd_with_org():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/contigs.fa', 'somename', 9, 'Salmonella')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"mkdir -p {args.prefix} && amrfinder -n {args.input} -o {args.prefix}/amrfinder.out --plus --organism {args.organism} --threads {args.jobs} -d {amr_obj.amrfinder_db}"
        amr_obj.logger = logging.getLogger()
        assert amr_obj._single_cmd() == cmd


def test_single_cmd_no_org():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/contigs.fa', 'somename', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"mkdir -p {args.prefix} && amrfinder -n {args.input} -o {args.prefix}/amrfinder.out --plus  --threads {args.jobs} -d {amr_obj.amrfinder_db}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._single_cmd() == cmd



def test_generate_cmd_single_fail():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/contigs.fa', 'somename', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"mkdir -p {args.prefix} && amrfinder -n {args.input} -o {args.prefix}/amrfinder.out -d {amr_obj.amrfinder_db}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._generate_cmd() != cmd


def test_generate_cmd_batch_fail():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("assembly", 'tests/batch.txt', '', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.amrfinder_db = "2021-06-01.1"
        amr_obj.identity = ''
        cmd = f"parallel -j {args.jobs} --colsep '\t' mkdir -p {{1}} && amrfinder -n {{2}} -o {{1}}/amrfinder.out -d {amr_obj.amrfinder_db} :::: {args.input}"
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._generate_cmd() != cmd


def test_check_output_file():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("assembly", 'tests/contigs.fa', 'tests', 9, '')
        amr_obj = RunFinder()
        p = 'tests/amrfinder.out'
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._check_output_file(p)


def test_check_output_file_fail():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("assembly", 'tests/contigs.fa', 'tests', 9, '')
        amr_obj = RunFinder()
        p = 'tests/amrfinders.out'
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.logger = logging.getLogger(__name__)
        with pytest.raises(SystemExit):
            amr_obj._check_output_file(p)


def test_check_outputs_single():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("assembly", 'tests/contigs.fa', 'tests', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._check_outputs()



def test_check_outputs_batch():
    """
    assert True when non-empty string is given
    """
    with patch.object(RunFinder, "__init__", lambda x: None):
        args = Data("batch", 'tests/batch.txt', 'tests', 9, '')
        amr_obj = RunFinder()
        amr_obj.organism = args.organism
        amr_obj.run_type = args.run_type
        amr_obj.prefix = args.prefix
        amr_obj.jobs = args.jobs
        amr_obj.input = args.input
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj._check_outputs()

# # test Collate
# # 
Colls = collections.namedtuple('Data', ['run_type', 'input', 'prefix'])
def test_get_drugclass_allele_1():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        df= pandas.read_csv('tests/amrfinder.out', sep = '\t')
        for i in df.iterrows():
            if i[1]['Gene symbol'] == 'blaSHV-11':
                row = i
        colname = 'allele'
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.get_drugclass(reftab, row, colname) == "Beta-lactam"

def test_extract_gene_name_1():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        protein = 'WP_004176269.1'
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.extract_gene_name(protein, reftab) == 'blaSHV-11'

def test_extract_gene_name_2():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        reftab = reftab.fillna('-')
        protein = 'WP_063839881.1'
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.extract_gene_name(protein, reftab) == "aac(2')-IIa"

def test_setup_dict():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        reftab = reftab.fillna('-')
        drugclass_dict = {}
        df= pandas.read_csv('tests/amrfinder.out', sep = '\t')
        for i in df.iterrows():
            if i[1]['Gene symbol'] == 'blaSHV-11':
                row = i
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.setup_dict(drugclass_dict, reftab, row) == {"Beta-lactam":['blaSHV-11']}


def test_get_per_isolate():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        reftab = reftab.fillna('-')
        
        df= pandas.read_csv('tests/amrfinder.out', sep = '\t')
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.get_per_isolate(reftab=reftab, df=df, isolate=isolate) == ({"Isolate":isolate, "Beta-lactam":'blaSHV-11','ESBL':'blaCTX-M-15'},{"Isolate":isolate,'ESBL':'blaCTX-M-15'},{"Isolate":isolate,'Metal':'qnrB1'})



def test_collate():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        reftab = pandas.read_csv(REFGENES)
        reftab = reftab.fillna('-')
        isolate = 'tests'
        drugs = pandas.DataFrame({"Isolate":isolate,'ESBL': 'blaCTX-M-15', "Beta-lactam":'blaSHV-11'}, index = [0])
        partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        virulence = pandas.DataFrame({"Isolate":isolate,'Metal':'qnrB1'}, index = [0])
        # print(drugs)
        
        amr_obj.logger = logging.getLogger(__name__)
        assert amr_obj.collate(isolate)[0].equals(drugs)
        assert amr_obj.collate(isolate)[1].equals(partial)
        assert amr_obj.collate(isolate)[2].equals(virulence)


def test_save():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'ESBL': 'blaCTX-M-15', "Beta-lactamase (not ESBL or carbapenemase)":'blaSHV-11'}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        virulence = pandas.DataFrame({"Isolate":isolate,'METAL':'qnrB1'}, index = [0])
        assert amr_obj.save_files('',summary_drugs,summary_partial, virulence)



def test_save_one_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        virulence = pandas.DataFrame({"Isolate":isolate,'METAL':'qnrB1'}, index = [0])
        assert amr_obj.save_files('',summary_drugs,summary_partial, virulence)



def test_save_all_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate}, index = [0])
        virulence = pandas.DataFrame({"Isolate":isolate}, index = [0])
        assert amr_obj.save_files('',summary_drugs,summary_partial, virulence)


def test_get_cols_works():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'ESBL': 'blaCTX-M-15', "Beta-lactamase (not ESBL or carbapenemase)":'blaSHV-11'}, index = [0])
        # summary_partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        # virulence = pandas.DataFrame({"Isolate":isolate,'METAL':'qnrB1'}, index = [0])
        assert amr_obj._get_cols(df= summary_drugs) == ["ESBL","Beta-lactamase (not ESBL or carbapenemase)"]

def test_get_cols_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame()
        # summary_partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        # virulence = pandas.DataFrame({"Isolate":isolate,'METAL':'qnrB1'}, index = [0])
        assert amr_obj._get_cols(df= summary_drugs) == []


def test_commbine_all_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame()
        summary_partial = pandas.DataFrame()
        virulence = pandas.DataFrame()
        assert amr_obj._combine_dfs(summary_drugs,summary_partial, virulence).empty



def test_save_merge():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'ESBL': 'blaCTX-M-15', "Beta-lactamase (not ESBL or carbapenemase)":'blaSHV-11'}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate,'ESBL':'blaCTX-M-15'}, index = [0])
        virulence = pandas.DataFrame()
        assert not amr_obj._merge(summary_drugs,summary_partial).empty



def test_save_merge_one_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'ESBL': 'blaCTX-M-15', "Beta-lactamase (not ESBL or carbapenemase)":'blaSHV-11'}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate}, index = [0])
        virulence = pandas.DataFrame()
        assert not amr_obj._merge(summary_drugs,summary_partial).empty



def test_save_merge_both_empty():
    """
    assert True when non-empty string is given
    """
    with patch.object(Collate, "__init__", lambda x: None):
        args = Colls("assembly", 'tests/contigs.fa', '')
        amr_obj = Collate()
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate}, index = [0])
        summary_partial = pandas.DataFrame({"Isolate":isolate}, index = [0])
        virulence = pandas.DataFrame()
        assert not amr_obj._merge(summary_drugs,summary_partial).empty



MduColls = collections.namedtuple('Data', ['sop', 'sop_name', 'qc','db','partials','matches','runid'])
def test_reporting_efaec_optra():

    with patch.object(Collate, "__init__", lambda x: None):
        args = MduColls('sop', 'sop_name', 'mduqc','db','partials','match','runid')
        amr_obj = MduCollate(args)
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'Florfenicol/Oxazolidinone': 'optrA'}, index = [0])
        
        for row in summary_drugs.iterrows():
            assert amr_obj.reporting_logic_general(row = row, species = 'Enterococcus faecium') == (['optrA'],["No non-reportable genes found."])


def test_reporting_ni_optra():

    with patch.object(Collate, "__init__", lambda x: None):
        args = MduColls('sop', 'sop_name', 'mduqc','db','partials','match','runid')
        amr_obj = MduCollate(args)
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        summary_drugs = pandas.DataFrame({"Isolate":isolate,'Florfenicol/Oxazolidinone': 'optrA'}, index = [0])
        
        for row in summary_drugs.iterrows():
            assert amr_obj.reporting_logic_general(row = row, species = 'No identification') == (['CPase_16S_mcr_NEG'],['optrA'])



def test_reporting_df_efaec():

    with patch.object(Collate, "__init__", lambda x: None):
        args = MduColls('sop', 'sop_name', f"{test_folder / 'mdu_qc_checked.csv'}",'db',f"{test_folder / 'summary_partials.txt'}",f"{test_folder / 'summary_matches.txt'}",'runid')
        amr_obj = MduCollate(args)
        isolate = 'tests'
        amr_obj.logger = logging.getLogger(__name__)
        
        df = pandas.DataFrame(
            {"MDU sample ID":isolate,
            'Item code':"",
            'Resistance genes (alleles) detected':"optrA",
            'Resistance genes (alleles) det (non-rpt)':"No non-reportable genes found.",
            'Species_obs':'Enterococcus faecium',
            'Species_exp':'No identification', 
            'db_version':'db'},
            index = [0]
        )
        df = df.set_index("MDU sample ID")
        
        assert amr_obj.mdu_reporting_general(match=f"{test_folder / 'summary_matches.txt'}").equals(df)



def test_reporting_df_shig():

    with patch.object(Collate, "__init__", lambda x: None):
        args = MduColls('sop', 'sop_name', f"{test_folder / 'mdu_qc_checked.csv'}",'db',f"{test_folder / 'summary_partials_shig.txt'}",f"{test_folder / 'summary_matches_shig.txt'}",'runid')
        amr_obj = MduCollate(args)
        isolate = 'tests2'
        amr_obj.logger = logging.getLogger(__name__)
        
        df = pandas.DataFrame(
            {"MDU sample ID":isolate,
            'Item code':"",
            'Resistance genes (alleles) detected':"CTX-M-15",
            'Resistance genes (alleles) det (non-rpt)':"No non-reportable genes found.",
            'Species_obs':'Eschericia coli', 
            'Species_exp':'Shigella flexneri',
            'db_version':'db'},
            index = [0]
        )
        df = df.set_index("MDU sample ID")
        
        assert amr_obj.mdu_reporting_general(match=f"{test_folder / 'summary_matches_shig.txt'}").equals(df)




import pandas, pathlib, numpy, subprocess,logging,json,datetime
from abritamr.CustomLog import CustomFormatter

logger =logging.getLogger(__name__) 
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(CustomFormatter())
fh = logging.FileHandler('update_abritamr_db.log')
fh.setLevel(logging.INFO)
formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
fh.setFormatter(formatter)
logger.addHandler(ch) 
logger.addHandler(fh)

def _get_date():

    return datetime.datetime.today().strftime('%Y-%m-%d')

def _get_keys(cfg):

    rename_key = cfg['rename_key']
    other_amr = cfg['other_amr']
    other_non_amr = cfg['other_non_amr']
    oxa_phen_list = cfg['oxa_phen_list']
    adrs = cfg['email_address']
    return rename_key,other_amr,other_non_amr,oxa_phen_list,adrs

def _get_vars():

    cfg = pathlib.Path(__file__).parent / "db" / "update_vars.json"
    if cfg.exists():
        with open(f"{cfg}", "r") as j:
            return _get_keys(json.load(j))
    else:
        logger.critical(f"It seems that {cfg} does not exist. Please check your installation and try again.")
        raise SystemExit
    
def _archive_old_ref_catalog():
    tdy = _get_date()
    og_ref = pathlib.Path(__file__).parent / "db" / "refgenes_latest.csv"
    archive = pathlib.Path(__file__).parent / "db" / f"refgenes_latest.csv.{tdy}"
    logger.info(f"Archiving reference catalog to : {archive}")
    subprocess.run(["cp", f"{og_ref}", f"{archive}"])

def _get_new_catalog():
    logger.info(f"Getting reference catalog from ncbi.")
    updated_html = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Antimicrobial_resistance/AMRFinderPlus/database/latest/ReferenceGeneCatalog.txt"
                    
    p = subprocess.run(f"wget -O ReferenceGeneCatalog.txt {updated_html}", shell = True, capture_output = True, encoding = "utf-8")
    if p.returncode == 0:
        logger.info(f"Reference catalog downloaded.")
        return True
    else:
        logger.critical(f"Something went wrong with the download of reference catalog. {p.stderr}.")
        raise SystemExit

def _make_key(df):
    if 'key' not in list(df.columns):
        df['mut_acc'] = df[['allele','whitelisted_taxa','refseq_nucleotide_accession']].apply( lambda x: '_'.join(x), axis = 1)
        df['key'] = numpy.where(df['refseq_protein_accession'] == '',df['genbank_protein_accession'],df['refseq_protein_accession'])
        df['key'] = numpy.where(df['subtype'] == 'POINT', df['mut_acc'], df['key'])

    return df

def _open_catalog():

    _get_new_catalog()
    logger.info(f"Generating a dataframe for update.")
    _new = pandas.read_csv("ReferenceGeneCatalog.txt", sep = "\t")
    _new = _new.fillna('')
    _new = _make_key(df = _new)
    return _new

def _get_previous_refgenes():

    existing = _check_existing()

    if existing != False:
        logger.info(f"An existing refgenes file has been found.")
        df = pandas.read_csv(existing)
        df = df.fillna('')
        return pandas.read_csv(existing)
    else:
        logger.warning(f"There is not an existing refgenes file. Refgenes file will be created from new.")
    return existing


def _check_existing():

    existing = pathlib.Path(__file__).parent /"db" / "refgenes_latest.csv"
    
    if existing.exists():
        return existing
    return False

def _capitalise(x):

    a = x.split('/')
    a = [i.capitalize() for i in a]
    _list = list(map(lambda x: x.replace('Carbapenem','Carbapenemase'), a))
 
    return '/'.join(_list)

def cfr(row):

    if row['gene_family'] == 'cfr' and row['class'] == '':
        # print(row)
        return 'Multidrug', _capitalise(row['subclass'])
    else:
        return _capitalise(row['class']),_capitalise(row['subclass'])


def _aminoglycosides(row):

    if 'rRNA' in row['product_name'] and 'methyltransferase' in row['product_name'] and 'AMINOGLYCOSIDE' in row['class']:

        return _capitalise(row['class']),'Aminoglycosides (Ribosomal methyltransferase)'

    else:
        return _capitalise(row['class']),_capitalise(row['subclass'])


def _oxa_phen(row):

    return 'Oxazolidinone/Phenicol',_capitalise(row['subclass'])

def _beta_lactams(row):
    
    if row['subtype'] == 'AMR-SUSCEPTIBLE':
        return 'Beta-lactam','Beta-lactam (Susceptible)'
    elif 'carbapenem-hydrolyzing' in row['product_name'] and 'OXA-51' not in row['product_name'] and row['subtype'] != 'POINT':
        # carbapenem-hydrolyzing
        return 'Beta-lactam','Carbapenemase'
    elif 'carbapenem-hydrolyzing' in row['product_name'] and 'OXA-51' not in row['product_name'] and row['subtype'] == 'POINT':
        # carbapenem-hydrolyzing
        return 'Beta-lactam','Carbapenemase'
    elif 'metallo-beta-lactamase' in row['product_name'] and row['subclass'] == 'CARBAPENEM':
        return 'Beta-lactam','Carbapenemase (MBL)'
    elif 'OXA-51 family' in row['product_name'] or row['allele'] == 'OXA-51':
        return 'Beta-lactam','Carbapenemase (OXA-51 family)'
    elif row['gene_family'] == 'blaZ':
        return 'Beta-lactam','Penicillin resistance (Staphylococcus aureus)'
    elif 'class C' in row['product_name']:
        return 'Beta-lactam','AmpC'
    elif row['gene_family'] == 'blaKPC' and row['subclass'] != 'CARBAPENEM':
        return 'Beta-lactam', 'ESBL (KPC variant)'
    # elif 'OXA-48 family' in row['product_name']:
    #     return 'Beta-lactam',_capitalise(row['subclass'])
    elif not 'class C' in row['product_name'] and ('extended-spectrum' in row['product_name'] or 'extended spectrum' in row['product_name']):
            return 'Beta-lactam','ESBL'
    else:
        return _capitalise(row['class']),_capitalise(row['subclass'])

def _other_antimicrobials(row):

    return 'Other antimicrobial', _capitalise(row['subclass'])

def _other_non_antimicrobials(row):

    return 'Other non-antimicrobial', _capitalise(row['subclass'])

def virulence(row):

    if row['class'] == 'INTIMIN':
        return 'Virulence', f"{_capitalise(row['class'])}_{row['subclass'].lower()}"
    elif 'STX' in row['class']:
        return 'Virulence', _capitalise(row['subclass'])
    elif row['class'] == '':
        return 'Virulence', 'Other'
    else:
        return 'Virulence', _capitalise(row['subclass'])

def _rename(rename_key, row):

    cls = _capitalise(row['class']) if row['class'] not in ['FLUOROQUINOLONE',"NITROFURAN"] else rename_key[row['class']]
    sbcls = rename_key[row['subclass']] if row['subclass'] in rename_key else _capitalise(row['subclass'])

    return cls,sbcls

def _logic(_dict,other_amr,other_non_amr,rename_key):
    logger.info(f"Applying classification logic.")
    for row in _dict:
        if row['class'] == 'AMINOGLYCOSIDE':
            row['enhanced_class'],row['enhanced_subclass'] = _aminoglycosides(row)
        # elif row['class'] == 'PHENICOL/OXAZOLIDINONE' or row['gene_family'] in oxa_phen_list:
        #     row['enhanced_class'],row['enhanced_subclass'] = _oxa_phen(row)
        elif row['gene_family'].startswith('cfr'):
            row['enhanced_class'],row['enhanced_subclass'] = cfr(row)
        elif row['class'] == 'BETA-LACTAM':
            row['enhanced_class'],row['enhanced_subclass'] = _beta_lactams(row)
        elif row['class'] in other_amr:
            row['enhanced_class'],row['enhanced_subclass'] = _other_antimicrobials(row)
        elif row['type'] == 'VIRULENCE':
            row['enhanced_class'],row['enhanced_subclass'] = virulence(row)
        elif row['class'] in other_non_amr:
            row['enhanced_class'],row['enhanced_subclass'] = _other_non_antimicrobials(row)
        elif row['class'] in rename_key:
            row['enhanced_class'],row['enhanced_subclass'] = _rename(rename_key=rename_key, row=row)
        elif row['class'] == 'MULTIDRUG':
            row['enhanced_class'],row['enhanced_subclass'] = 'Multidrug',_capitalise(row['subclass'])
        elif 'multidrug' in row['product_name'] and  (row['class'] == '' and row['subclass'] == ''):
            row['enhanced_class'],row['enhanced_subclass'] = 'Multidrug','Other'
        elif row['type'] != 'VIRULENCE' and (row['class'] == '' and row['subclass'] == ''):
            row['enhanced_class'],row['enhanced_subclass'] = 'Other','Other'
        else:
            row['enhanced_class'],row['enhanced_subclass'] = _capitalise(row['class']),_capitalise(row['subclass'])
    return _dict

def _make_dict(df,other_amr,other_non_amr,rename_key):
    _new_dict = df.to_dict(orient='records')
    _new_dict = _logic(_dict = _new_dict,other_amr =other_amr,other_non_amr=other_non_amr,rename_key=rename_key)
    
    return _new_dict

def _email(adrs,pth):
    # f"cat {_email} | mailx -s '{runid} QC' {a} {user_email}"

    cmd = f"echo 'Please confirm curation of the abritAMR refgenes database attached here.' | mailx -s 'abritamr update' -a {pth} {adrs}"
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding='utf-8')

    if p.returncode == 0:
        logger.info("abritAMR update email sent for approval.")
    else:
        logger.critical(f"Something has gone wrong. {p.stderr}.")

def _save_df(df):
    
    sfx = _get_date()
    df.to_csv(f"refgenes_{sfx}.csv", index = False)
    return f"refgenes_{sfx}.csv"

def _updated_entries(new_catalog, previous_catalog):
    # print(previous_catalog)
    previous_catalog = previous_catalog.rename(columns = {"class_new": "class_prev", "subclass_new":"subclass_prev"})
    new_catalog = new_catalog.rename(columns = {"class":"class_new", "subclass": "subclass_new"})
    _tmp= new_catalog.merge(previous_catalog, on = ['key'])
    print(_tmp.columns)
    _tmp['changed'] = numpy.where((_tmp['class_prev'] != _tmp['class_new']) , 'updated', '')
    _tmp['changed'] = numpy.where((_tmp['subclass_prev'] != _tmp['subclass_new']) , 'updated', _tmp['changed'])
    
    _tmp['Previous_class'] = numpy.where(_tmp['changed'] == 'updated', _tmp['class_prev'], '')
    _tmp['Previous_subclass'] = numpy.where(_tmp['changed'] == 'updated', _tmp['subclass_prev'],'')
    changed = list(_tmp[_tmp['changed']!='']['key'])
    
    new_catalog['Status'] = numpy.where(new_catalog['key'].isin(changed), 'updated',new_catalog['Status'])
    new_catalog = new_catalog.merge(_tmp[['key','Previous_class','Previous_subclass']], on = ['key'], how = 'left')
    return new_catalog

def _new_entries(new_catalog, previous_catalog):
    
    logger.info(f"Checking for new entries.")
    new_catalog['Status'] = numpy.where(new_catalog['key'].isin(list(previous_catalog['key'])), 'existing','new')
    logger.info(f"Checking for updated entries.")
    new_catalog = _updated_entries(new_catalog=new_catalog,previous_catalog=previous_catalog)
    return new_catalog

def _update_status(new_catalog, previous_catalog):

    new_catalog = _new_entries(new_catalog=new_catalog,previous_catalog=previous_catalog)

    return new_catalog

def _compare_to_existing(new_catalog, previous_catalog):

    tab = pandas.DataFrame(new_catalog)
    
    if isinstance(previous_catalog,pandas.DataFrame):
        new_catalog = _update_status(new_catalog=tab,previous_catalog=previous_catalog)
        return new_catalog
    else:
        logger.info(f"There is in no previous file to compare with.")
        return tab
    

def create_refgenes():

    rename_key,other_amr,other_non_amr,oxa_phen_list,adrs = _get_vars()
    new_catalog_df = _open_catalog()
    new_catalog = _make_dict(df = new_catalog_df ,other_amr =other_amr,other_non_amr=other_non_amr,rename_key=rename_key)
    existing = _get_previous_refgenes()
    new_catalog = _compare_to_existing(new_catalog=new_catalog,previous_catalog=existing)
    pth = _save_df(df=new_catalog)
    logger.info(f"Saved new refgenes catalog as {pth} and emailing for confirmation.")
    _email(adrs = adrs, pth = pth)
    



#!/usr/bin/env python3
import pathlib, pandas, math, sys,  re, logging, numpy
import warnings
pandas.options.mode.chained_assignment = None
# from pandas.core.algorithms import isin
from abritamr.CustomLog import CustomFormatter

class Collate:

    """
    a base class for collation of amrfinder results - to be used when not doing MDU QC
    """
    
    ANNOTATIONS = {'blast':'*','partial':'^','exact':''}
    REFGENES = pathlib.Path(__file__).parent / "db" / "refgenes_latest.csv"
    MATCH = ["ALLELEX", "BLASTX", "EXACTX", "POINTX"]

    def __init__(self, args):
        self.logger =logging.getLogger(__name__) 
        self.logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(CustomFormatter())
        fh = logging.FileHandler('abritamr.log')
        fh.setLevel(logging.INFO)
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
        fh.setFormatter(formatter)
        self.logger.addHandler(ch) 
        self.logger.addHandler(fh)
        self.prefix = args.prefix
        self.run_type = args.run_type
        self.input = args.input

    def joins(self, dict_for_joining):
        """
        make them a comma separated list
        """
        for i in dict_for_joining:
            if i != "Isolate":
                dict_for_joining[i] = sorted(list(set(dict_for_joining[i])))
                dict_for_joining[i] = ",".join(dict_for_joining[i])

        return dict_for_joining

    def get_drugclass(self, reftab, row, colname):

        """
        if the enhanced subclass is in either NONRTM or MACROLIDES then then use the groups specified by Norelle. If it is empty (-) then fall back on the AMRFinder subclass, else report the extended subclass
        """
        gene_id_col = "Gene symbol" if colname != "refseq_protein_accession" else "Accession of closest sequence" # to get the name of drug and the drugclass
        
        if reftab[reftab[colname] == row[1][gene_id_col]].empty:
            d = 'Unknown'
            for i in ['genbank_protein_accession','refseq_nucleotide_accession','genbank_nucleotide_accession']:
                # print(reftab[reftab[i] == row[1][gene_id_col]])
                if not reftab[reftab[i] == row[1][gene_id_col]].empty:
                    d= reftab[reftab[i] == row[1][gene_id_col]]['enhanced_subclass'].values[0]
                    break
        else:
            
            d = reftab[reftab[colname] == row[1][gene_id_col]]['enhanced_subclass'].values[0]
        
        return d

    def extract_bifunctional_name(self, protein, reftab):
        """
        extract the joint name of bifunctional genes
        """
        return reftab[reftab["refseq_protein_accession"] == protein]["gene_family"].values[0]

    def extract_gene_name(self, protein, reftab, pointn = False):
        
        # add in to address the calling of point mutations that are not an exact match to reference
        # amrfinderplus returns the nucleotide accession with the range for what was detected.. these
        # so extract just the name of accession and set col to work with to nucleotide rather than protein
        # Add POINTN to end so users know this is a different type of match 
        suff = f"_POINTN" if pointn else ""
        protein = protein.split(':')[0] if pointn else protein
        col ="refseq_nucleotide_accession" if pointn else "refseq_protein_accession"
        
        try:
            if reftab[reftab[col] == protein]["allele"].values[0] != '-':
                nme = f"{reftab[reftab[col] == protein]['allele'].values[0]}{suff}"
                return nme
            else:
                return reftab[reftab[col] == protein]["gene_family"].values[0]
        except IndexError:
            for i in ['genbank_protein_accession','refseq_nucleotide_accession','genbank_nucleotide_accession']:
                if len(reftab[reftab[i] == protein]["gene_family"].unique()) >= 1:
                    return reftab[reftab[i] == protein]["gene_family"].unique()[0]
            
    def setup_dict(self, drugclass_dict, reftab, row, _type = 'exact', pointn = False):
        """
        return the dictionary for collation
        """
        if row[1]["Gene symbol"] in list(reftab["allele"]) and 'POINT' not in row[1]['Method']:
            drugclass = self.get_drugclass(
                    reftab=reftab, row=row, colname="allele"
                    )
            drugname = self.extract_gene_name(protein = row[1]["Accession of closest sequence"], reftab = reftab, pointn = pointn)
        elif row[1]["Gene symbol"] in list(reftab["allele"]) and 'POINT' in row[1]['Method']:
            drugclass = self.get_drugclass(
                    reftab=reftab, row=row, colname="allele"
                    )
            drugname = row[1]["Gene symbol"]
            
        elif row[1]["Gene symbol"] in list(reftab["gene_family"]):
            
            drugclass = self.get_drugclass(
                reftab=reftab, row=row, colname="refseq_protein_accession"
            )
            drugname = f"{self.extract_gene_name(protein = row[1]['Accession of closest sequence'], reftab = reftab)}*" if not row[1]["Method"] in ["EXACTX", "ALLELEX"] else f"{self.extract_gene_name(protein = row[1]['Accession of closest sequence'], reftab = reftab)}"
            
        elif row[1]["Accession of closest sequence"] in list(reftab["refseq_protein_accession"]):
            
            drugclass = self.get_drugclass(
                reftab = reftab, row = row, colname = "refseq_protein_accession"
            )
            drugname = self.extract_bifunctional_name(protein = row[1]['Accession of closest sequence'], reftab = reftab)
        else:
            drugname = row[1]["Gene symbol"]
            drugclass = "Unknown"

        if drugclass in drugclass_dict:
            drugclass_dict[drugclass].append(drugname)
        elif drugclass not in drugclass_dict:
            drugclass_dict[drugclass] = [drugname]

        return drugclass_dict

    def _other_dict(self, other_dict, row):
        """
        report virulence and stress genes
        """
        if row[1]['Element subtype'].capitalize() in other_dict:
            other_dict[row[1]['Element subtype'].capitalize()].append(row[1]['Gene symbol'])
        else:
            other_dict[row[1]['Element subtype'].capitalize()] = [row[1]['Gene symbol']]
        return other_dict

    def get_per_isolate(self, reftab, df, isolate):
        """
        make three dictionaries for each isolate that contain the drug class assignments for each match that is one of ALLELEX,POINTX, EXACTX or BLASTX, another dictionary which lists all partial mathces and a dictionary of virulence factors
        """
        drugclass_dict = {"Isolate": isolate}
        partials = {"Isolate": isolate}
        other = {"Isolate": isolate}
        for row in df.iterrows():
            # if the match is good then generate a drugclass dict
            if row[1]["Gene symbol"] == "aac(6')-Ib-cr" and row[1]["Method"] in ["EXACTX", "ALLELEX"]: # This is always a partial - unclear
                partials = self.setup_dict(drugclass_dict = partials, reftab = reftab, row = row)
            elif row[1]["Method"] in self.MATCH and row[1]["Element type"] == "AMR" and row[1]['Element subtype'] != "AMR-SUSCEPTIBLE":
                drugclass_dict = self.setup_dict(drugclass_dict = drugclass_dict, reftab = reftab, row = row)
            elif "POINTN" in row[1]["Method"] and row[1]["Element type"] == "AMR" and row[1]['Element subtype'] != "AMR-SUSCEPTIBLE":
                drugclass_dict = self.setup_dict(drugclass_dict = drugclass_dict, reftab = reftab, row = row, pointn = True)
            elif row[1]["Method"] not in self.MATCH and row[1]["Element type"] == "AMR" and row[1]['Element subtype'] != "AMR-SUSCEPTIBLE":
                partials = self.setup_dict(drugclass_dict = partials, reftab = reftab, row = row)
            else:
                other = self._other_dict(other_dict = other, row = row)
        drugclass_dict = self.joins(dict_for_joining=drugclass_dict)
        partials = self.joins(dict_for_joining=partials)
        other = self.joins(dict_for_joining = other)
        return drugclass_dict, partials, other

    def _get_cols(self, df, _not = 'Isolate'):
        
        cols = [c for c in df.columns if c != _not]

        return cols

    def _add_caret(self,df,cols):

        for c in cols:
            df[c] = numpy.where(df[c]!="", df[c].str.strip("*") + "^", df[c])
        return df

    def _merge(self, match, partial):

        df = pandas.DataFrame()
        tmp = match.merge(partial, on = 'Isolate', how = 'outer')
        tmp_cols = self._get_cols(df = tmp)
        
        for t in tmp_cols:
            if t.endswith('_x') or t.endswith('_y'): 
                # if this is a a col that has appeared in more than one df
                if df.empty:
                    df = tmp[['Isolate']]
                c = t.split('_')[0] # strip the _x or _y for col name
                if c not in list(df.columns): # if this has not already been added - should never happen but best safe than sorry
                    df[c] = tmp[[f"{c}_x", f"{c}_y"]].apply(lambda x: ','.join([i for i in x if i != '']), axis = 1) # combine the values and join with a comma
            else:
                # if this is a column that only appears in one df
                if df.empty:
                    df = tmp[['Isolate', t]]
                else:
                    df = df.merge(tmp[['Isolate', t]], on = 'Isolate',how = 'outer') 
        
        return df if not df.empty else tmp

    def _combine_dfs(self, match, partial, virulence):
        
        match = match.fillna('')
        partial = partial.fillna('')
        pcols = self._get_cols(df = partial)
        mcols = self._get_cols(df = match)
        vcols = self._get_cols(df = virulence)
        
        if vcols == [] and pcols == [] and mcols == []:
            return pandas.DataFrame()

        if pcols != []:
            partial = self._add_caret(df = partial, cols = pcols)      
        
        df = self._merge(match = match, partial = partial)
        
        if not isinstance(virulence, pandas.DataFrame) and not df.empty:
            df = df.merge(virulence, on = 'Isolate', how = 'outer') 
        elif isinstance(virulence, pandas.DataFrame) and df.empty:
            if not virulence.empty:
                df = virulence
            else:
                df = pandas.DataFrame()

        return df

    def save_files(self, path, match, partial, virulence):
        
        files = {'summary_matches.txt': match, 'summary_partials.txt': partial, 'summary_virulence.txt':virulence}
        
        for f in files:
            out = f"{path}/{f}" if path != '' else f"{f}"
            self.logger.info(f"Saving {out}")
            files[f].set_index('Isolate').to_csv(f"{out}", sep = '\t')
        combd = self._combine_dfs(match = match, partial = partial, virulence = virulence)
        combd_out = f"{path}/abritamr.txt" if path != '' else f"abritamr.txt"
        if not combd.empty:
            self.logger.info(f"Saving combined file : {combd_out}")
            combd.set_index('Isolate').to_csv(f"{combd_out}", sep = '\t')
        
        return True
        

    def _get_reftab(self):
        """
        get reftab
        """

        reftab = pandas.read_csv(self.REFGENES)
        reftab = reftab.fillna("-")

        return reftab

    def collate(self, prefix = ''):
        """
        if the refgenes.csv is present then proceed to collate data and save the csv files.
        """

        
        reftab = self._get_reftab()
        
        df = pandas.read_csv(f"{prefix}/amrfinder.out", sep="\t")
        self.logger.info(f"Opened amrfinder output for {prefix}")
        drug, partial, virulence = self.get_per_isolate(
            reftab=reftab, df=df, isolate=prefix
        )
        
        summary_drugs = pandas.DataFrame(drug, index = [0])
        summary_partial = pandas.DataFrame(partial, index = [0])
        summary_virulence = pandas.DataFrame(virulence, index = [0])
        return summary_drugs, summary_partial,summary_virulence
        
    
    def _combine_df(self, existing, temp):
        """
        combine result dataframes for batch
        """
        if existing.empty:
            existing = temp
        else:
            existing = pandas.concat([existing,temp])
        
        return existing

    def _batch_collate(self,input_file):


        summary_matches = pandas.DataFrame()
        summary_partial = pandas.DataFrame()
        summary_virulence = pandas.DataFrame()

        df = pandas.read_csv(input_file, sep = '\t', header = None)
        for row in df.iterrows():
            prefix = f"{row[1][0]}"
            self.logger.info(f"Collating results for {prefix}")
            temp_match, temp_partial, temp_virulence = self.collate(prefix = prefix)
            summary_matches = self._combine_df(existing = summary_matches, temp = temp_match)
            summary_partial = self._combine_df(existing = summary_partial, temp = temp_partial)
            summary_virulence = self._combine_df(existing = summary_virulence, temp = temp_virulence)
        
        return summary_matches, summary_partial, summary_virulence

    def run(self):


        if not pathlib.Path(self.REFGENES).exists():
            self.logger.critical(f"The refgenes DB ({self.REFGENES}) seems to be missing.")
            raise SystemExit

        if self.run_type != 'batch':
            self.logger.info(f"This is a single sample run.")
            summary_drugs, summary_partial, virulence = self.collate(prefix = self.prefix)
        else:
            self.logger.info(f"You are running abritamr in batch mode. Your collated results will be saved.")
            summary_drugs, summary_partial, virulence = self._batch_collate(input_file = self.input)
        self.logger.info(f"Saving files now.")
        self.save_files(path='' if self.run_type == 'batch' else f"{self.prefix}", match = summary_drugs,partial=summary_partial, virulence = virulence)
        
class MduCollate(Collate):
    
    def __init__(self, args):
        self.logger =logging.getLogger(__name__) 
        self.logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(CustomFormatter())
        fh = logging.FileHandler('abritamr.log')
        fh.setLevel(logging.INFO)
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p') 
        fh.setFormatter(formatter)
        self.logger.addHandler(ch) 
        self.logger.addHandler(fh)
        self.sop = args.sop
        self.sop_name = args.sop_name
        self.mduqc = args.qc
        self.db = args.db
        self.partials = args.partials
        self.match = args.matches
        self.runid = args.runid
        self.NONE_CODES = {
            "Salmonella":"CPase_ESBL_AmpC_16S_NEG",
            "Shigella":"CPase_ESBL_AmpC_16S_NEG",
            "Staphylococcus":"Mec_VanAB_Linez_NEG",
            "Enterococcus":"Van_Linez_NEG",
            "Other":"CPase_16S_mcr_NEG"
        }
        self.REPORTING = {"Salmonella enterica":self.mdu_reporting_salmonella}

    def mdu_qc_tab(self):
        self.logger.info(f"Checking the format of the QC file")
        cols = ["ISOLATE", 'SPECIES_EXP', 'SPECIES_OBS', 'TEST_QC']
        tab = pandas.read_csv(self.mduqc)
        tab = tab.rename(columns = {tab.columns[0]: 'ISOLATE'})
        for c in cols:
            if c not in list(tab.columns):
                self.logger.critical(f"There seems to be a problem with your QC file. This file must have {','.join(cols)}. Please check your input and try again.")
                raise SystemExit

        pos = pandas.DataFrame(data = {"ISOLATE": "9999-99888", "TEST_QC" : "PASS", "SPECIES_EXP":"Staphylococcus aureus", "SPECIES_OBS":"Staphylococcus aureus" }, index = [0])
        
        return pandas.concat([tab,pos])

    def strip_bla(self, gene):
        '''
        strip bla from front of genes except
        '''
        if gene.startswith("bla") and len(gene) >6 and gene.endswith("*"):
            gene = gene.replace("bla", "")
        elif gene.startswith("bla") and len(gene) >5 and not gene.endswith("*"):
            gene = gene.replace("bla", "")
        return gene


    # Carbapenemase to be reported for all cases
    # Carbapenemase (MBL) all in all HOWEVER if blaL1 should  not be reported in Stenotrophomonas maltophilia
    # Carbapenemase (OXA-51 family) REPORTED IN ALL except in Acinetobacter baumannii,Acinetobacter calcoaceticus,Acinetobacter nosocomialis,Acinetobacter pittii,Acinetobacter baumannii complex,
    # All ESBL in Salmonella and Shigella
    # ESBL (Amp C type) in Salmonella and Shigella
    # Aminoglycosides (Ribosomal methyltransferases) ALL
    # Colistin ALL
    # Oxazolidinone & phenicol resistance if Genus = Enterococcus or Staphylococcus aureus and Staphylococcus argenteus
    # report vanA*, B*, C*, vanD*, vanE*, vanG*, vanL*, vanM*, vanN* 
    # Methicillin ALL

    def get_all_genes(self, row):
        all_genes = []
        for r in row[1]:
            if isinstance(r, str):
                if len(r.split(",")) > 1:
                    for a in r.split(","):
                        all_genes.append(a)
                else:
                    all_genes.append(r)
        return all_genes

    def none_replacement_code(self, genus):

        if genus in self.NONE_CODES:
            return self.NONE_CODES[genus]
        else:
            return "CPase_16S_mcr_NEG"

    def assign_itemcode(self,mduid, reg):
        self.logger.info(f"Checking for item code")
        m = reg.match(mduid)
        try:
            itemcode = m.group('itemcode') if m.group('itemcode') else ''
        except AttributeError:
            itemcode = ''
        return itemcode

    def assign_mduid(self, mduid, reg):
        self.logger.info(f"Extracting MDU sample ID")
        m = reg.match(mduid)
        try:
            mduid = m.group('id')
        except AttributeError:
            mduid = mduid.split('/')[-1]
        return mduid

    def _ampicillin_res_sal(self, col, gene):
        
        if col in [ 'Beta-lactam','ESBL','AmpC'] or 'Ampicillin' in col:
            return gene
        return ''

    def _chloramphenicol_res_sal(self, col, gene):

        if 'phenicol' in col.lower():
            return gene
        return ''

    def _cefo_esbl_res_sal(self, col, gene):

        if col == 'ESBL':
            return gene
        return ''

    def _cefo_ampc_res_sal(self, col, gene):

        if 'AmpC' in col :
            return gene
        return ''
    
    def _carbapenem_res_salmo(self, col, gene):

        if "Carbapenemase" in col: #not KPC
            return gene
        return ''

    def _azi_res_salmo(self, col, gene):

        if 'Azithromycin' in  col or col in ['Macrolide', 'Lincosamide/Macrolide/Streptogramin']:
            return gene
        return ''

    def _gentamicin_res_salm(self, col, gene):
        if col in ['Gentamicin', 'Aminoglycosides (Ribosomal methyltransferase)']: 
            return gene
        return ''
    
    def _kanamycin_res_salm(self, col, gene):
    
        if col in ['Kanamycin','Aminoglycosides (Ribosomal methyltransferase)']:
            return gene
        return ''
    
    def _streptomycin_res_salm(self, col, gene):
        if 'Streptomycin' in col:
            return gene
        return ''
    
    def _spectinomycin_res_salm(self, col, gene):
        if 'Streptomycin' in col:
            return gene
        return ''
    
    def _tetra_res_salmo(self, col, gene):

        if col == 'Tetracycline':
            return gene
        return ''

    def _cipro_res_salmo(self, col, gene):

        if 'Quinolone' in col:
            return gene
        return ''

    def _sulf_res_salmo(self, col, gene):
        
        if col == "Sulfonamide":
            return gene
        return ''
    
    def _trimet_res_salmo(self, col, gene):

        if col == "Trimethoprim":
            return gene
        return ''
    
    def _trim_sulpha_salmo(self, trim_gene, sul_gene):

        return f"{trim_gene},{sul_gene}"

    def _rmt_res_salmo(self, col, gene):

        if col == "Aminoglycosides (Ribosomal methyltransferase)":
            return gene
        return ''

    def _colistin_res_salmo(self, col, gene):
        if col == 'Colistin':
            return gene
        return ''
    
    def reporting_logic_salmonella(self, row):

        # Ampicillin - ResMech	Ampicillin - Interpretation	
        # Cefotaxime (ESBL) - ResMech	Cefotaxime (ESBL) - Interpretation	
        # Cefotaxime (AmpC) - ResMech	Cefotaxime (AmpC) - Interpretation	
        # Tetracycline - ResMech	Tetracycline - Interpretation	
        # Gentamycin - ResMech	Gentamycin - Interpretation	
        # Sulfathiazole - ResMech	Sulfathiazole - Interpretation	
        # Trimethoprim - ResMech	Trimethoprim - Interpretation	
        # Ciprofloxacin - ResMech	Ciprofloxacin - Interpretation	
        # Azithromycin - ResMech	Azithromycin - Interpretation

        mduidreg = re.compile(r'(?P<id>[0-9]{4}-[0-9]{5,6})-?(?P<itemcode>.{1,})?')
        
        all_genes = self.get_all_genes(row)
        all_genes = [a for a in all_genes if a != row[1]['Isolate'] and a != '']
                
        isodict = row[1].to_dict()
        item_code = self.assign_itemcode(row[1]['Isolate'], mduidreg)
        md = self.assign_mduid(row[1]['Isolate'], mduidreg)

        abx = {
            "Ampicillin" : self._ampicillin_res_sal,
            "Cefotaxime (ESBL)":self._cefo_esbl_res_sal,
            "Cefotaxime (AmpC)":self._cefo_ampc_res_sal,
            "Meropenem" :self._carbapenem_res_salmo,
            "Azithromycin":self._azi_res_salmo,
            "Gentamicin":self._gentamicin_res_salm,
            "Kanamycin":self._kanamycin_res_salm,
            "Streptomycin":self._streptomycin_res_salm,
            "Spectinomycin":self._spectinomycin_res_salm,
            "Tetracycline":self._tetra_res_salmo,
            "Ciprofloxacin":self._cipro_res_salmo,
            "Sulfathiazole":self._sulf_res_salmo,
            "Trimethoprim":self._trimet_res_salmo,
            "Trim-Sulpha":self._trim_sulpha_salmo,
            "Chloramphenicol":self._chloramphenicol_res_sal,
            "Aminoglycosides (RMT)": self._rmt_res_salmo,
            "Colistin":self._colistin_res_salmo
        }

        
        tmp_results = {
            "Ampicillin":[],	
            "Cefotaxime (ESBL)":[],
            "Cefotaxime (AmpC)":[],
            "Tetracycline":[],
            "Gentamicin":[],
            "Sulfathiazole":[],
            "Trimethoprim":[],
            "Trim-Sulpha":[],
            "Kanamycin":[],
            "Streptomycin":[],
            "Spectinomycin":[],
            "Ciprofloxacin":[],
            "Azithromycin":[],
            "Meropenem":[],
            "Chloramphenicol":[],
            "Aminoglycosides (RMT)":[],"Colistin":[], "Other":[]
        }

        
        for ab in abx:
            if ab != 'Trim-Sulpha':
                for i in isodict:
                    if i != "Isolate":
                        
                        g = abx[ab](col = i, gene = isodict[i])
                        
                        if g != '':
                            gene_list = g.split(',')
                            for gene in gene_list:
                                tmp_results[ab].append(gene)
                                if gene in all_genes:
                                    all_genes.remove(gene)
            
        # interprete results
        tmp_results['Other'] = all_genes
        results = {'Isolate': row[1]['Isolate'], 'MDU Sample ID':md, 'Item code': item_code}
        # for Trim-Sulpha
        
        if tmp_results['Trimethoprim'] != [] and tmp_results['Sulfathiazole'] != []:
            tmp_results['Trim-Sulpha'] = list(set(tmp_results['Trimethoprim']).union(tmp_results['Sulfathiazole']))
        
        for res in tmp_results:
            results[f"{res} - ResMech"] = ';'.join(tmp_results[res]) if tmp_results[res] != [] else "None detected"              
            if res in ["Aminoglycosides (RMT)","Colistin", "Other"]:
                results[f"{res} - Interpretation"] = ''
            elif res == 'Ciprofloxacin':
                if tmp_results[res] == []:
                    results[f"{res} - Interpretation"] = 'Susceptible'
                # elif len(tmp_results[res]) == 1:
                #     results[f"{res} - Interpretation"] = 'Intermediate'
                else:
                    results[f"{res} - Interpretation"] = 'Resistant'
            elif res not in ["Aminoglycosides (RMT)","Colistin", "Other"]:
                if tmp_results[res] == []:
                    results[f"{res} - Interpretation"] = 'Susceptible'
                else:
                    results[f"{res} - Interpretation"] = 'Resistant'
        
        return results

            
    def reporting_logic_general(self, row, species, neg_code = True):
        
        # get all genes found
        all_genes = self.get_all_genes(row)
        isodict = row[1].to_dict()
        # determine the genus EXPECTED
        genus = species.split()[0]
        reportable = [
            "Carbapenemase",
            "Carbapenemase (MBL)",
            "ESBL (KPC variant)"
            "Carbapenemase (OXA-51 family)",
            "ESBL",
            "AmpC", # will need to change to reflect new subclass AmpC
            "Aminoglycosides (Ribosomal methyltransferase)",
            "Colistin",
            "Chloramphenicol/Florfenicol/Linezolid", 
            "Florfenicol/Oxazolidinone",
            "Vancomycin",
            "Methicillin"
        ]
        non_caveat_reportable = [
            "Carbapenemase",
            "ESBL (KPC variant)",
            "Aminoglycosides (Ribosomal methyltransferase)",
            "Colistin"
        ]
        # Carbapenemase (KPC variant)
# Carbapenemase (MBL)
# Carbapenemase (OXA-51 family)

        abacter_excluded = [
            "Acinetobacter baumannii",
            "Acinetobacter calcoaceticus",
            "Acinetobacter nosocomialis",
            "Acinetobacter pittii",
            "Acinetobacter baumannii complex"
        ]

        
        van_match = re.compile(r"van[A,B,C,D,E,G,L,M,N][\S]*")
        mec_match = re.compile(r"mec[^IR]")
        
        genes_reported = []  # genes for reporting
        genes_not_reported = []  # genes found but not reportable
        for i in isodict:
            genes = []
            if i != 'Isolate':
                if isinstance(isodict[i], str):
                    genes = isodict[i].split(',')
                    genes = [gene for gene in genes if '_' not in gene] # this is to remove the point mutations for MMS118
                if genes != []: # for each bin we do things to genes
                    if i in reportable: #don't report point mutations
                        # print(i)
                        if i in non_caveat_reportable:
                            genes_reported.extend(genes)
                        elif i == "Carbapenemase (MBL)" and species != "Stenotrophomonas maltophilia":
                            genes_reported.extend(genes)
                        elif i == "Carbapenemase (MBL)" and species == "Stenotrophomonas maltophilia":
                            # if species is "Stenotrophomonas maltophilia" don't report blaL1
                            genes_reported.extend([g for g in genes if not g.startswith("blaL1")])
                            genes_not_reported.extend([g for g in genes if g.startswith("blaL1")])
                        elif i == "Carbapenemase (OXA-51 family)" and species not in abacter_excluded:
                            genes_reported.extend(genes)
                        elif i in ["ESBL","AmpC"] and genus in ["Salmonella"]: 
                            genes_reported.extend(genes)
                        elif i in ["ESBL","AmpC"] and genus in ["Shigella"]: 
                            genes_reported.extend([g for g in genes if "blaEC" not in g])
                            genes_not_reported.extend([g for g in genes if "blaEC" in g]) # don't report blaEC for shigella
                        elif i == "Vancomycin":
                            genes_reported.extend([g for g in genes if van_match.match(g)])
                            genes_not_reported.extend([g for g in genes if not van_match.match(g)])
                        elif i == "Methicillin":
                            genes_reported.extend([g for g in genes if mec_match.match(g)])
                            genes_not_reported.extend([g for g in genes if not mec_match.match(g)])
                        
                        elif "Oxazolidinone" in i or "Linezolid" in i:
                            if species in ["Staphylococcus aureus","Staphylococcus argenteus"] or genus == "Enterococcus":
                                genes_reported.extend(genes)
                            else:
                                genes_not_reported.extend(genes)
                        else:
                            genes_not_reported.extend(genes)

                    else:
                        genes_not_reported.extend(genes)
        if genes_reported == []:
            genes_reported = [self.none_replacement_code(genus= genus)] if neg_code else ''
        if genes_not_reported == []:
            genes_not_reported = ["No non-reportable genes found."] if neg_code else ''
            # break
        
        self.logger.info(f"{row[1]['Isolate']} has {len(genes_reported)} reportable genes.")
        return genes_reported, genes_not_reported

    def mdu_reporting_salmonella(self, match, isolates):

        self.logger.info(f"Applying MDU business logic for interpretation of  Salmonella AST")
        cols = ["MDU Sample ID", "Item code", 
        "Ampicillin - ResMech", "Ampicillin - Interpretation",
        "Cefotaxime (ESBL) - ResMech","Cefotaxime (ESBL) - Interpretation",
        "Cefotaxime (AmpC) - ResMech","Cefotaxime (AmpC) - Interpretation",
        "Tetracycline - ResMech","Tetracycline - Interpretation",
        "Gentamicin - ResMech","Gentamicin - Interpretation",
        "Kanamycin - ResMech",	"Kanamycin - Interpretation",
        "Streptomycin - ResMech",	"Streptomycin - Interpretation",
        "Sulfathiazole - ResMech","Sulfathiazole - Interpretation",
        "Trimethoprim - ResMech","Trimethoprim - Interpretation",
        "Trim-Sulpha - ResMech",	"Trim-Sulpha - Interpretation",
        "Chloramphenicol - ResMech",	"Chloramphenicol - Interpretation",
        "Ciprofloxacin - ResMech","Ciprofloxacin - Interpretation",
        "Meropenem - ResMech",	"Meropenem - Interpretation",
        "Azithromycin - ResMech","Azithromycin - Interpretation",
        "Aminoglycosides (RMT) - ResMech",
        "Aminoglycosides (RMT) - Interpretation",
        "Colistin - ResMech",
        "Colistin - Interpretation", 
        "Other - ResMech",
        "Other - Interpretation"]
        # select passed Salmonella
        
        df = pandas.read_csv(match, sep = '\t')
        df = df[df['Isolate'].isin(isolates)]
        result_df = pandas.DataFrame()
        df = df.fillna('')
        for row in df.iterrows():
            isolate_results = self.reporting_logic_salmonella(row = row)
            tmpdf = pandas.DataFrame(isolate_results, index = [0])
            if result_df.empty:
                result_df = tmpdf
            else:
                result_df = result_df.append(tmpdf)
        
        return result_df[cols]

    def _extract_plus_isolates(self,species):

        qc = self.mdu_qc_tab()
        qc = qc[(qc['SPECIES_OBS'] == species) & (qc['TEST_QC'] == 'PASS')]

        return list(qc["ISOLATE"])


    def mdu_reporting_general(self, match, neg_code = True):

        self.logger.info(f"Applying MDU business logic {'matches' if neg_code else 'partials'}.")
        mduidreg = re.compile(r'(?P<id>[0-9]{4}-[0-9]{5,6})-?(?P<itemcode>.{1,})?')
        reporting_df = pandas.DataFrame()
        qc = self.mdu_qc_tab()
        match_df = pandas.read_csv(match, sep = '\t')
        for row in match_df.iterrows():
            isolate = row[1]['Isolate']
            item_code = self.assign_itemcode(isolate, mduidreg)
            md = self.assign_mduid(isolate, mduidreg)
            d = {"MDU sample ID": md, "Item code" : item_code}
            qcdf = qc[qc['ISOLATE'].str.contains(isolate)]
            exp_species = qcdf["SPECIES_EXP"].values[0]
            obs_species = qcdf["SPECIES_OBS"].values[0]
            if obs_species == exp_species:
                species = exp_species
            elif obs_species != exp_species and 'Shigella' in exp_species:
                species = exp_species
            else:
                species = obs_species
            # species = obs_species if obs_species == exp_species else obs_species
            genes_reported, genes_not_reported = self.reporting_logic_general(
                row=row, species=species, neg_code=neg_code
            )
            
            # strip bla
            genes_not_reported = [self.strip_bla(g) for g in genes_not_reported]
            genes_not_reported = [g for g in genes_not_reported if g != isolate]
            genes_reported = [self.strip_bla(g) for g in genes_reported]
            genes_reported = [g for g in genes_reported if g != isolate]
            d["Resistance genes (alleles) detected"] = ",".join(genes_reported)
            d["Resistance genes (alleles) det (non-rpt)"] = ",".join(genes_not_reported)
            if qcdf["TEST_QC"].values[0] == 'FAIL': # species not needed for MDU LIMS upload
                d["Species_exp"] = exp_species
            d["Species_obs"] = obs_species
            d["Species_exp"] = exp_species
            d['db_version'] = self.db
            tempdf = pandas.DataFrame(d, index=[0])
            tempdf = tempdf.set_index("MDU sample ID")
            
            if reporting_df.empty:
                reporting_df = tempdf
            else:
                reporting_df = pandas.concat([reporting_df,tempdf])
            
        return reporting_df.reindex(labels = ['Item code','Resistance genes (alleles) detected','Resistance genes (alleles) det (non-rpt)','Species_obs', 'Species_exp', 'db_version'], axis = 'columns')

    
    def save_spreadsheet_general(
        self,
        passed_match,
        passed_partials,
        
    ):
        self.logger.info(f"Saving {self.sop_name}.")
        writer = pandas.ExcelWriter(f"{self.runid}_{self.sop_name}.xlsx", engine="xlsxwriter")
        passed_match.to_excel(writer, sheet_name=f"{self.sop_name}")
        passed_partials.to_excel(writer, sheet_name="Passed QC partial")
        writer.close()

    def save_spreadsheet_interpreted(self, results):
        sheets = {"Salmonella enterica":f"{self.sop_name}-01"}
        self.logger.info(f"Saving MMS184")
        writer = pandas.ExcelWriter(f"{self.runid}_{self.sop_name}.xlsx", engine = "xlsxwriter")
        for result in results:
            result[1].to_excel(writer, sheet_name = sheets[result[0]], index = False)
        writer.close()

    def run(self):
        if self.sop == 'general' and pathlib.Path(self.partials).exists():
            passed_match_df = self.mdu_reporting_general(match=self.match)
            passed_partials_df = self.mdu_reporting_general(match = self.partials)
            self.save_spreadsheet_general(
                passed_match_df,
                passed_partials_df
            )
        elif self.sop == 'plus':
            dfs = []
            for r in self.REPORTING:
                self.logger.info(f"Checking if there are any {r} in this run.")
                isolates = self._extract_plus_isolates(species = r)
                if isolates != []:
                    self.logger.info(f"There are {len(isolates)} {r} in this run.")
                    plus_df = self.mdu_reporting_salmonella(match = self.match, isolates=isolates)
                    dfs.append((r,plus_df))
                else:
                    self.logger.info(f"There are no {r} in this run. Collation will be skipped.")
            self.save_spreadsheet_interpreted(results = dfs)
